import logging
from collections import defaultdict
from copy import deepcopy

from django.conf import settings
from django.core.cache import cache
from django.db import transaction
from rest_framework.exceptions import PermissionDenied
from rest_framework.request import Request

from ..models import User
from ..models.user_role import ROLES, UserRole
from ..signals import pre_update_idp_user, post_update_idp_user
from ..typing import UserUpdateEvent, UserRecordDict
from ..utils.functions import get_or_none, keep_keys, update_record, cache_user_service_results

logger = logging.getLogger(__name__)

APP_IDENTIFIER = settings.IDP_USER_APP['APP_IDENTIFIER']
IN_DEV = settings.APP_ENV == "development"


class UserService:

    # Service Methods Used by Django Application
    @staticmethod
    def get_role(request: Request):
        return request.query_params.get('role')

    @staticmethod
    def authorize_resources_or_get_all_authorized_resources(
            user: User,
            role: ROLES,
            resource: str,
            resource_ids: list[int],
            permission: str = None
    ):
        if resource_ids:
            UserService.authorize_resources(
                user=user,
                role=role,
                resource=resource,
                resource_ids=resource_ids,
                permission=permission
            )
            return resource_ids
        else:
            return UserService.get_authorized_resources(
                user=user,
                role=role,
                resource=resource,
                permission=permission
            )

    @staticmethod
    def authorize_resources(user: User, role: ROLES, resource: str, resource_ids: list[int], permission: str = None):
        """
        It specifies if the user is authorized to access the objects/permissions that he is requesting
        Permissions that do have restrictable=True, are allowed from OPA, so we need to solve this on product level
        Args:
            user:           The user trying to access the resources
            resource:       Resource name, could be originator_instance_ids; vehicle_ids; portfolio_ids,
                                depending on how it was set up in IDP.
            resource_ids:   The ids of the resources within the configuration
            permission:     In case of specific permissions such as DodManager we can have
                            permission restrictions through IDP. The value is the name of the permission
            role:           ROLES as defined in the settings file.

        Returns: None; In case the requested objects are not accessible  it will raise errors instead.
        """

        user_role = UserRole.objects.get(user=user, role=role)

        if permission:
            permission_restrictions = user_role.permission_restrictions
            if permission_restrictions and permission in permission_restrictions.keys():
                permission_restriction = permission_restrictions.get(permission)
                restriction_value = permission_restriction.get(resource)
                if restriction_value is False:
                    raise PermissionDenied('You are not allowed to access the resources in the Requested Objects!')
                else:
                    if not set(resource_ids).issubset(set(restriction_value)):
                        raise PermissionDenied('You are not allowed to access the resources in the Requested Objects!')
                    else:
                        return

        # Check App config
        if not set(resource_ids).issubset(set(user_role.app_config.get(resource, {}))):
            raise PermissionDenied(f'You are not allowed to access the resources in the Requested Objects!')

    @staticmethod
    @cache_user_service_results
    def get_authorized_resources(user: User, role, resource: str, permission: str = None) -> list[int]:
        """
        It gets the authorized resources for the group that has been requested.
        Args:
            user:           Logged in User
            role:           ROLES as defined in the settings file.
            resource:       The resource being accessed. For example, vehicle_ids
            permission:     In case of specific permissions such as DodManager we can have
                            permission restrictions through IDP. The value is the name of the permission

        Returns:
            Resource Ids
        """
        assert ROLES.as_dict().get(role) is not None, f"Role does not exist: {role}"

        try:
            user_role = UserRole.objects.get(user=user, role=role)

            if permission:
                permission_restrictions = user_role.permission_restrictions
                if permission_restrictions and permission in permission_restrictions.keys():
                    permission_restriction = permission_restrictions.get(permission)
                    return permission_restriction.get(resource) or []

            return user_role.app_config.get(resource, [])
        except UserRole.DoesNotExist:
            return []

    @staticmethod
    def _create_or_update_user(data: UserUpdateEvent) -> User:
        user = get_or_none(User.objects, idp_user_id=data.get("idp_user_id"))
        user_data = keep_keys(data, [
            "idp_user_id",
            "username",
            "email",
            "first_name",
            "last_name",
        ])
        if user:
            update_record(user, **user_data)
            UserService._invalidate_user_cache_entries(user=user)
            return user
        else:
            return User.objects.create(**user_data)

    @staticmethod
    def _invalidate_user_cache_entries(user: User):
        """
        Invalidate all the entries in the cache for the given user.
        To do this, find all the entries that start with the app identifier and username of the user.
        """
        if settings.IDP_USER_APP.get('USE_REDIS_CACHE', False) and not IN_DEV:
            cache.delete_pattern(f"{APP_IDENTIFIER}-{user.username}*")

    @classmethod
    def update_user(cls, data: UserRecordDict):
        """
        Extract tenants from the user record and call _update_user for each tenant.
        Remove tenant information from the payload of _update_user since it is not needed
        inside of it.

        Send signals before and after calling the _update_user method for each tenant
        separately. This gives possibility to the project to react on the user update.

        One case of handling this signal is to switch database connection to the tenant's
        database. In this way the user can be updated in the correct database.

        """
        reported_user_app_configs = UserService._get_reported_user_app_configs(data)
        tenants = reported_user_app_configs.keys()

        for tenant in tenants:
            logger.info(f"Updating user {data['username']} for tenant {tenant}")
            pre_update_idp_user.send(sender=cls.__class__, tenant=tenant)

            user_record_for_tenant = deepcopy(data)
            user_record_for_tenant['app_specific_configs'] = reported_user_app_configs[tenant]

            with transaction.atomic(using=tenant):
                UserService._update_user(user_record_for_tenant)

            post_update_idp_user.send(sender=cls.__class__, tenant=tenant)

    @staticmethod
    def _update_user(data: UserUpdateEvent):
        """
        This method makes sure that the changes that are coming from the IDP
        for a user are propagated in the internal product Authorization Schemas

        Step 1: Create or update User Object
        Step 2: Create/Update/Delete User Roles for this user.
        """

        user = UserService._create_or_update_user(data)

        current_user_roles = defaultdict()
        for user_role in user.user_roles.all():
            current_user_roles[user_role.role] = user_role

        reported_user_app_configs = UserService._get_reported_user_app_configs(data)

        for role, role_data in reported_user_app_configs.items():
            if existing_user_role := current_user_roles.get(role):
                update_record(
                    existing_user_role,
                    permission_restrictions=role_data.get('permission_restrictions'),
                    app_config=role_data.get("app_config")
                )
            else:
                UserRole.objects.create(
                    user=user,
                    role=role,
                    permission_restrictions=role_data.get('permission_restrictions'),
                    app_config=role_data.get("app_config")
                )

        # Verify if any of the previous user roles is not being reported anymore
        # Delete it if this is the case
        for role, user_role in current_user_roles.items():  # type: str, UserRole
            if reported_user_app_configs.get(role) is None:
                user_role.delete()

    @staticmethod
    def _get_reported_user_app_configs(data):
        return data.get('app_specific_configs', {}).get(APP_IDENTIFIER, {})



import json
import logging
from json import JSONDecodeError

import requests
from django.conf import settings
from requests import Response
from rest_framework import permissions
from rest_framework.request import Request

from idp_user.services import OpaService

APP_IDENTIFIER = settings.IDP_USER_APP["APP_IDENTIFIER"]

logger = logging.getLogger(__name__)


class OpaCheckPermission(permissions.BasePermission):
    message = 'Opa has blocked you from accessing the resource!'

    @classmethod
    def _get_opa_response(cls, request: Request) -> Response:
        opa_domain = settings.IDP_USER_APP['OPA_DOMAIN']
        opa_version = settings.IDP_USER_APP['OPA_VERSION']
        url = f"{opa_domain}/{opa_version}/data/{APP_IDENTIFIER}/{settings.APP_ENV}/allow"

        role = request.GET.get('role')
        roles_functionalities = json.loads(cls._get_request_header(request, 'X-ROLES-FUNCTIONALITIES'))
        allowed_functionalities = roles_functionalities.get(role, [])

        request_body = {
            'input': {
                'allowed_functionalities': allowed_functionalities,
                'path': request.path,
                'method': request.method,
                'role': role
            }
        }

        # Ask OPA for a policy decision
        return requests.post(url, json=request_body)

    @staticmethod
    def _get_request_header(request: Request, header: str) -> str:
        return request.headers.get(header) or request.META.get(header)

    def has_permission(self, request, view):

        logger.info(f"OPA: Checking permission for {view}!")
        response = self._get_opa_response(request)
        if not response.ok:
            logger.info(f"OPA: Cannot Authorize at the moment!")
            return False
        try:
            response_body = response.json()
        except JSONDecodeError:
            logger.info(f"OPA: Decode Error, cannot Authorize at the moment!")
            return False

        try:
            allow = response_body['result']
            if not allow:
                logger.info(f"OPA: Forbidden Resource!")
                return False
            logger.info(f"OPA: Resource is granted!")
            return True
        except KeyError:
            # If the key result is not present in the response,
            # it means the policy for this app does not exist in OPA
            # Update OPA with the necessary policy and data in this case
            logger.info("Opa: Policy/Data is missing, instructing IDP to recreate the rules")
            OpaService.update_opa(authorization_header=request.headers.get('Authorization'))
            # Reattempt authorize
            logger.info("OPA: Re-attempting Permissions")
            return self.has_permission(request, view)



class DoesNotExist(Exception):
    def __init__(self, message="Record does not exist!"):
        super().__init__(message)


class AuthenticationError(Exception):
    def __init__(self, detail: str = "Authentication Error"):
        super().__init__(detail)


class MissingHeaderError(Exception):
    def __init__(self, header_name):
        super().__init__(f"Missing header: {header_name}")



import json

from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned


APP_IDENTIFIER = settings.IDP_USER_APP['APP_IDENTIFIER']
IN_DEV = settings.APP_ENV == "development"


def keep_keys(dictionary, keys):
    return {k: v for k, v in dictionary.items() if k in keys}


def get_or_none(records, *args, **kwargs):
    try:
        return records.get(*args, **kwargs)
    except ObjectDoesNotExist:
        return None
    except MultipleObjectsReturned:
        return records.filter(*args, **kwargs)[0]


def update_record(record, save=True, **data):
    if data:
        for key, value in data.items():
            setattr(record, key, value)
        if save:
            record.save()
    return record


def cache_user_service_results(function):
    def wrapper(user, *args, **kwargs):
        cache_key = f"{APP_IDENTIFIER}-{user.username}-{function.__name__}"
        for arg in args:
            cache_key += f",{arg}"
        for key, value in kwargs.items():
            cache_key += f",{key}={value}"

        result = cache.get(cache_key)
        if result:
            return json.loads(result)
        else:
            result = function(user=user, *args, **kwargs)
            cache.set(cache_key, json.dumps(result))
            return result

    if IN_DEV or not settings.IDP_USER_APP.get('USE_REDIS_CACHE', False):
        return function

    return wrapper



import logging
import os
from typing import Optional, Callable, Literal, List

import jwt
import requests
from django.conf import settings
from rest_framework import authentication
from rest_framework.request import Request

from idp_user.models import User
from idp_user.typing import JwtData
from idp_user.utils.exceptions import AuthenticationError, MissingHeaderError

logger = logging.getLogger(__name__)

APP_IDENTIFIER = settings.IDP_USER_APP["APP_IDENTIFIER"]

# Allow header injection only if in Development Environment, for testing purposes
INJECT_HEADERS = settings.IDP_USER_APP.get("INJECT_HEADERS_IN_DEV", False) and settings.APP_ENV == 'development'


class AuthenticationBackend(authentication.TokenAuthentication):

    @classmethod
    def _auth_required(cls, request: Request):
        cls._access_token_required(request)
        cls._ingress_headers_required(request)

    @classmethod
    def _access_token_required(cls, request: Request):
        try:
            cls._header_required(request, 'Authorization', 'token', lambda x: x.startswith('Bearer '))
        except MissingHeaderError:
            raise AuthenticationError("Missing Token.")

    @classmethod
    def _ingress_headers_required(cls, request: Request):
        cls._header_required(request, 'X-USER-ID', 'User ID', lambda x: x.isdigit())
        cls._header_required(request, 'X-ROLES-FUNCTIONALITIES', 'roles functionalities')

    @classmethod
    def _header_required(
            cls,
            request: Request,
            header: str,
            header_human_name: str,
            validation_func: Optional[Callable] = None,
    ):
        header = cls._get_request_header(request, header)
        if not header:
            raise MissingHeaderError(header_human_name)
        if validation_func and not validation_func(header):
            raise MissingHeaderError(header_human_name)

    @classmethod
    def _verify_user_id(cls, request: Request, jwt_data: JwtData):
        if not int(cls._get_request_header(request, 'X-USER-ID')) == jwt_data['user_id']:
            raise AuthenticationError('Invalid token.')

    @classmethod
    def _get_access_token(cls, request: Request) -> str:
        auth_header = cls._get_request_header(request, 'Authorization')
        return auth_header.replace('Bearer ', '')

    @classmethod
    def _get_jwt_payload(cls, request: Request) -> JwtData:
        try:
            # Signature already verified from ingress before
            data = jwt.decode(cls._get_access_token(request), algorithms=['HS256'], options={'verify_signature': False})
        except Exception:
            raise AuthenticationError('Invalid token.')
        return data

    @classmethod
    def _verify_jwt_claims(cls, jwt_data: JwtData, claims: List[Literal['user_id']]):
        for claim in claims:
            cls._verify_jwt_claim(jwt_data, claim)

    @classmethod
    def _verify_jwt_claim(cls, jwt_data: JwtData, claim: Literal['user_id']):
        try:
            jwt_data[claim]
        except KeyError:
            raise AuthenticationError('Invalid token.')

    @staticmethod
    def _get_request_header(request: Request, header: str) -> str:
        return request.headers.get(header) or request.META.get(header)

    @staticmethod
    def _get_user(user_id: int) -> User:
        try:
            return User.objects.get(idp_user_id=user_id)
        except User.DoesNotExist:
            pass

    @staticmethod
    def __inject_headers_through_idp(request: Request):
        response = requests.get(
            url=f"{os.getenv('IDP_URL')}/api/validate/?app={APP_IDENTIFIER}",
            headers={
                "Authorization": request.headers.get('Authorization'),
            }
        )
        if response.ok:
            # Headers dict cannot be modified, insert the headers as META,
            # that will be used by the overriden method _get_request_header
            request.META['X-USER-ID'] = response.headers.get('X-USER-ID')
            request.META['X-ROLES-FUNCTIONALITIES'] = response.headers.get('X-ROLES-FUNCTIONALITIES')
            return request

    def __skip_auth_headers_and_opa(self, request: Request):
        try:
            # Require access token
            self._access_token_required(request)
            # Get jwt payload from access token
            jwt_data = self._get_jwt_payload(request)
            # Check if required user claims are provided
            self._verify_jwt_claims(jwt_data, ['user_id'])

            return self._get_user(jwt_data['user_id']), self

        except AuthenticationError:
            return None, None

    def authenticate(self, request: Request):
        try:
            # Check if ingress headers are provided
            self._ingress_headers_required(request)

            self._access_token_required(request)

            jwt_data = self._get_jwt_payload(request)
            # Check if required user claims are provided
            self._verify_jwt_claims(jwt_data, ['user_id'])
            # Check if user id added by ingress in the request is the same as the user id in the token
            self._verify_user_id(request, jwt_data)
            # Set the user in the request for later access
            return self._get_user(jwt_data['user_id']), self

        except MissingHeaderError:
            if INJECT_HEADERS:
                request = self.__inject_headers_through_idp(request)
                if not request:
                    return None, None
                return self.authenticate(request)
            else:
                return self.__skip_auth_headers_and_opa(request)
        except AuthenticationError:
            return None, None



from drf_spectacular.extensions import OpenApiAuthenticationExtension
from drf_spectacular.openapi import AutoSchema
from drf_spectacular.utils import OpenApiParameter

from idp_user.auth.authentication import AuthenticationBackend


class BearerTokenScheme(OpenApiAuthenticationExtension):
    target_class = AuthenticationBackend  # full import path OR class ref
    name = 'IDPAuthentication'  # name used in the schema

    def get_security_definition(self, auto_schema):
        return {
            "type": "http",
            "scheme": "bearer",
            "bearerFormat": "JWT",
        }


class AutoSchemaWithRole(AutoSchema):
    def get_override_parameters(self):
        return [
            OpenApiParameter("role", type=str, location=OpenApiParameter.QUERY, required=True),
        ]



from django.conf import settings
from django.db import models

from django.utils.module_loading import import_string

ROLES = import_string(settings.IDP_USER_APP.get('ROLES'))


class UserRole(models.Model):
    user = models.ForeignKey(to="idp_user.User", related_name="user_roles", on_delete=models.CASCADE)
    role = models.CharField(choices=ROLES.as_list(), max_length=125)
    # This dictionary contains explicit restrictions regarding vehicle related features, in the form:
    # {"ViewDod": {"vehicle_ids": [1,2]}, "synchronizationDoD": False}
    app_config = models.JSONField(null=True)
    permission_restrictions = models.JSONField(default=dict)

    class Meta:
        unique_together = [('user', 'role')]



# Generated by Django 3.1.13 on 2021-10-26 09:12

import django.contrib.auth.models
import django.contrib.auth.validators
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
    ]

    operations = [
        migrations.CreateModel(
            name='User',
            fields=[
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('is_superuser', models.BooleanField(default=False,
                                                     help_text='Designates that this user has all permissions without explicitly assigning them.',
                                                     verbose_name='superuser status')),
                ('username', models.CharField(error_messages={'unique': 'A user with that username already exists.'},
                                              help_text='Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.',
                                              max_length=150, unique=True,
                                              validators=[django.contrib.auth.validators.UnicodeUsernameValidator()],
                                              verbose_name='username')),
                ('first_name', models.CharField(blank=True, max_length=150, verbose_name='first name')),
                ('last_name', models.CharField(blank=True, max_length=150, verbose_name='last name')),
                ('email', models.EmailField(blank=True, max_length=254, verbose_name='email address')),
                ('is_staff', models.BooleanField(default=False,
                                                 help_text='Designates whether the user can log into this admin site.',
                                                 verbose_name='staff status')),
                ('is_active', models.BooleanField(default=True,
                                                  help_text='Designates whether this user should be treated as active. Unselect this instead of deleting accounts.',
                                                  verbose_name='active')),
                ('date_joined', models.DateTimeField(default=django.utils.timezone.now, verbose_name='date joined')),
                ('idp_user_id', models.BigIntegerField(primary_key=True, serialize=False)),
                ('groups', models.ManyToManyField(blank=True,
                                                  help_text='The groups this user belongs to. A user will get all permissions granted to each of their groups.',
                                                  related_name='user_set', related_query_name='user', to='auth.Group',
                                                  verbose_name='groups')),
                ('user_permissions', models.ManyToManyField(blank=True, help_text='Specific permissions for this user.',
                                                            related_name='user_set', related_query_name='user',
                                                            to='auth.Permission', verbose_name='user permissions')),
            ],
            options={
                'verbose_name': 'user',
                'verbose_name_plural': 'users',
                'abstract': False,
            },
            managers=[
                ('objects', django.contrib.auth.models.UserManager()),
            ],
        ),
        migrations.CreateModel(
            name='UserRole',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('role', models.CharField(max_length=125)),
                ('app_config', models.JSONField(null=True)),
                ('permission_restrictions', models.JSONField(default=dict)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='user_roles',
                                           to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'unique_together': {('user', 'role')},
            },
        ),
    ]



from django.conf import settings
from django.core.checks import Warning, Error, register


@register()
def check_idp_user_settings(*args, **kwargs):
    def verify_idp_user_app_attr_exists(_attr, error=False):
        if _attr not in idp_user_settings:
            issue_kwargs = {
                'msg': f'Missing {_attr} in IDP_USER_APP.',
                'hint': f'You must declare the attribute {_attr} in the dict IDP_USER_APP in settings.py.',
                'obj': settings,
                'id': f'idp_user.{_attr}',
            }
            if error:
                issues.append(Error(**issue_kwargs))
            else:
                issues.append(Warning(**issue_kwargs))

    issues = []

    try:
        settings.APP_ENV
    except AttributeError:
        issues.append(
            Error(
                'Missing APP_ENV.',
                hint='You must declare the APP_ENV (development/staging/production) variable in settings.py.',
                obj=settings,
                id='idp_user.E001',
            )
        )

    try:
        auth_user_model = settings.AUTH_USER_MODEL
        if auth_user_model != 'idp_user.User':
            issues.append(
                Warning(
                    'Wrong AUTH_USER_MODEL.',
                    hint="You must set AUTH_USER_MODEL = 'idp_user.User' in settings.py.",
                    obj=settings,
                    id='idp_user.E002',
                )
            )

    except AttributeError:
        issues.append(
            Warning(
                'Missing AUTH_USER_MODEL',
                hint="You must set AUTH_USER_MODEL = 'idp_user.user' in settings.py.",
                obj=settings,
                id='idp_user.E003',
            )
        )

    try:
        idp_user_settings = settings.IDP_USER_APP

        for attr in ["APP_IDENTIFIER", "FAUST_APP_PATH", "OPA_DOMAIN", "OPA_VERSION"]:
            verify_idp_user_app_attr_exists(attr, error=True)

        for attr in ["ROLES", "IDP_URL"]:
            verify_idp_user_app_attr_exists(attr, error=False)

    except AttributeError:
        issues.append(
            Error(
                'Missing IDP_USER_APP.',
                hint='You must declare the idp_user settings in the variable IDP_USER_APP in settings.py.',
                obj=settings,
                id='idp_user.E004',
            )
        )

    return issues



from typing import TypedDict, Union, List, Any, Optional


class JwtData(TypedDict):
    iat: int
    nbf: int
    jti: str
    exp: str
    type: str
    fresh: str
    user_id: int
    email: str
    username: str


class UserFeaturesPermissions(TypedDict):
    dod_manager: Union[List, bool]
    cash_flow_projection: Union[List, bool]
    notes_manager: Union[List, bool]


class AppSpecificConfigs(TypedDict):
    app_config: Any
    permission_restrictions: dict[str, Union[bool, Any]]


Role = str
UserAppSpecificConfigs = dict[Role, AppSpecificConfigs]


class UserUpdateEvent(TypedDict):
    idp_user_id: int
    first_name: str
    last_name: str
    username: str
    email: str
    app_specific_configs: UserAppSpecificConfigs


"""
"data": [
   {
        "idp_user_id": 12,
        "first_name": "str",
        "last_name": "str",
        "username": "str",
        "email": "str",
        "app_specific_configs": {
            "app_identifier": {
                "Servicer": {
                    "app_config": {"vehicle_ids": [1, 2]},
                    "permission_restrictions": {
                        "viewDoD": {"vehicle_ids": [1]},
                        "synchronizeDoD": false
                    }
                }
            }
        }
    }
]
"""

# ===
AppIdentifier = str
TenantIdentifier = str

UserRecordAppSpecificConfigs = dict[AppIdentifier, dict[TenantIdentifier, AppSpecificConfigs]]


class UserRecordDict(TypedDict):
    idp_user_id: int
    first_name: Optional[str]
    last_name: Optional[str]
    username: Optional[str]
    email: Optional[str]
    app_specific_configs: UserRecordAppSpecificConfigs


"""
Example of a user record from kafka:
{
    "first_name": "str",
    "last_name": "str",
    "username": "str",
    "email": "str",
    "app_specific_configs": {
        "app_identifier": {
            "tenant": {
                "Servicer": {
                    "app_config": {"vehicle_ids": [1, 2]},
                    "permission_restrictions": {
                        "viewDoD": {"vehicle_ids": [1]},
                        "synchronizeDoD": false
                    }
                }
            }
        }
    }
}
"""
